"""A Store is any output channel from a rule."""
from abc import abstractmethod

from sinai import BaseView
from sinai.exceptions import MetricNotFound
from sinai.types import JDict, MetricId, MetricInstance, MetricResult


class Store(BaseView):
    """A class for storing data.
    This is a base class, it doesn't do much."""

    ...

    @abstractmethod
    def save_metric(self, metric: MetricInstance) -> None:
        """Save a metric instance into the store.
        If this is not a metric store then don't implement this method.
        """
        ...


class MetricStore(Store):
    """A class for storing metrics.
    This is a base class, for reference, it doesn't do much either."""

    def save_metric(self, metric: MetricInstance) -> None:
        """Save a metric instance into the store."""
        if metric.update:
            self._upsert(metric)
        else:
            self._save(metric)

    def _upsert(self, metric: MetricInstance) -> None:
        try:
            existing, metric_id = self._get_metric_for_update(metric)
        except MetricNotFound:
            self._save(metric)
        else:
            self._replace(existing, metric_id)

    def _get_metric_for_update(self, metric: MetricInstance) -> MetricResult:
        metric_filter: JDict = {}
        for field in metric.update:
            metric_filter[field] = getattr(metric, field)
        return self._find_metric_by_filter(metric, metric_filter)

    @abstractmethod
    def _save(self, metric: MetricInstance) -> None:
        ...

    @abstractmethod
    def _find_metric_by_filter(
        self, metric: MetricInstance, metric_filter: JDict
    ) -> MetricResult:
        ...

    @abstractmethod
    def _replace(self, metric: MetricInstance, metric_id: MetricId) -> None:
        ...



from uuid import uuid4

from sinai.adaptors import Memory
from sinai.exceptions import MetricNotFound
from sinai.stores import MetricStore
from sinai.types import JDict, MetricId, MetricInstance, MetricResult


class MemoryMetricStore(MetricStore, Memory):
    """Basic in memory metric store.
    Use a database-backed store to get persistence."""

    def _save(self, metric: MetricInstance) -> None:
        self.memory[uuid4()] = metric.pre_save(self.monitor.id)

    def _replace(self, metric: MetricInstance, metric_id: MetricId) -> None:
        self.memory[metric_id] = metric.pre_save(self.monitor.id)

    def _find_metric_by_filter(
        self, metric: MetricInstance, metric_filter: JDict
    ) -> MetricResult:
        if results := self.find_metric(metric_filter):
            for metric_id, existing_metric in results.items():
                metric.created_at = existing_metric.created_at
                # This could support updating multiple metrics, for now support only the first
                return metric, metric_id
        raise MetricNotFound(f"Metric with fields: {metric_filter} not found.")



from statistics import mean, median, mode

from sinai.metrics.base import Metric
from sinai.types import AggregationFunction, MetricValueList


class AggregationMetric(Metric):
    """A metric formed from summarizing a query of stored metrics."""

    name = "aggregation"
    aggregation: AggregationFunction = len

    def process_data_points(self, data_points: MetricValueList) -> None:
        """Turn multiple values into one."""
        self.value = self.aggregation(data_points)  # type:ignore


class CountMetric(AggregationMetric):
    """The total number of matching metrics."""

    name = "count"
    aggregation = len


class SumMetric(AggregationMetric):
    """The total values of matching metric."""

    name = "sum"
    aggregation = sum  # type:ignore


class MeanMetric(AggregationMetric):
    """The average of matching metric values."""

    name = "mean"
    aggregation = mean


class MaxMetric(AggregationMetric):
    """The maximum of the matching metric values."""

    name = "max"
    aggregation = max  # type:ignore


class MinMetric(AggregationMetric):
    """The minimum of the matching metric values."""

    name = "min"
    aggregation = min  # type:ignore


class MedianMetric(AggregationMetric):
    """The median of the matching metric values."""

    name = "median"
    aggregation = median


class ModeMetric(AggregationMetric):
    """The mode of the matching metric values."""

    name = "mode"
    aggregation = mode


AGGREGATION_CLASSES = {
    "count": CountMetric,
    "sum": SumMetric,
    "mean": MeanMetric,
    "max": MaxMetric,
    "min": MinMetric,
    "median": MedianMetric,
    "mode": ModeMetric,
}



"""In Memory Store."""

from sinai.metrics import Metric
from sinai.types import GlobalMemory, JDict, MetricDict


class Memory:
    """A simple global memory dict.
    For persistence, use a database instead.
    """

    memory: GlobalMemory = {}

    def find_metric(self, metric_filter: JDict) -> MetricDict:
        """Find a dictionary in the in-memory store."""
        results = {}
        for metric_id, metric_dict in self.memory.items():
            for filter_key in metric_filter:
                if metric_dict[filter_key] != metric_filter[filter_key]:
                    continue
                metric = Metric.from_dict(metric_dict)
                results[metric_id] = metric
        return results

    def clear(self) -> None:
        self.memory = {}



"""Support for using Mongo Databases as a source or store for data."""

from __future__ import annotations

try:
    import pymongo

except ImportError:
    HAS_MONGO = False
else:
    HAS_MONGO = True

from sinai.exceptions import DependencyMissing
from sinai.types import MonitorInstance


class Mongo:
    """A connection to MongoDB."""

    connection_string = ""
    database_name = ""

    def __init__(self, monitor: MonitorInstance) -> None:
        self.monitor = monitor
        if not HAS_MONGO:
            raise DependencyMissing(
                "Pymongo library required for using Mongo databases."
            )
        self.client = pymongo.MongoClient(self.connection_string)  # type: ignore
        self.db = self.client[self.database_name]
        self.collection = self.db["metrics"]



"""The Api adaptors provide sources and stores using API endpoints.
Requires the requests library to be installed.
"""
try:
    import requests
except ImportError:
    HAS_REQUESTS = False
else:
    HAS_REQUESTS = True

from types import ModuleType

from sinai import BaseView
from sinai.exceptions import DependencyMissing
from sinai.types import MonitorInstance, RequestHeader


class ApiConnection(BaseView):
    """Base class to share common connection information between API Stores and Sources"""

    url: str = ""
    headers: RequestHeader = {}
    token: str = ""

    def __init__(self, monitor: MonitorInstance):
        self.monitor = monitor
        super().__init__(monitor)
        if not HAS_REQUESTS:
            raise DependencyMissing("Requests library required for API calls.")
        self.requests: ModuleType = requests

    def get_headers(self) -> RequestHeader:
        """The required authentication, by default uses a bearer token."""
        if self.token:
            return self.bearer_headers(self.token)
        return self.headers

    @staticmethod
    def bearer_headers(token: str) -> RequestHeader:
        """Headers for APIs with bearer token authentication."""
        return {
            "accept": "application/json",
            "Authorization": "Bearer " + token,
        }



__all__ = [
    "Metric",
    "AggregationMetric",
    "CountMetric",
    "SumMetric",
    "MeanMetric",
    "MaxMetric",
    "MinMetric",
    "MedianMetric",
    "ModeMetric",
    "AGGREGATION_CLASSES",
]

from sinai.metrics.aggregation import (
    AGGREGATION_CLASSES,
    AggregationMetric,
    CountMetric,
    MaxMetric,
    MeanMetric,
    MedianMetric,
    MinMetric,
    ModeMetric,
    SumMetric,
)
from sinai.metrics.base import Metric



"""Retrieving data and metics from MongoDB."""

from sinai.adaptors.mongo import Mongo
from sinai.metrics import Metric
from sinai.sources import MetricSource, Source
from sinai.types import (
    FList,
    JDict,
    JDictOrNone,
    MetricInstance,
    MetricList,
    MongoResult,
)


class MongoSource(Mongo, Source):
    """Gets arbitrary documents from MongoDB."""

    def find_one(
        self,
        collection: str,
        mongo_filter: JDictOrNone = None,
        *args: FList,
        **kwargs: JDict,
    ) -> MongoResult:
        """Find a single document from MongoFB."""
        return self.db[collection].find_one(mongo_filter, *args, **kwargs)

    def find(
        self,
        collection: str,
        mongo_filter: JDictOrNone = None,
        *args: FList,
        **kwargs: JDict,
    ) -> MongoResult:
        """Find documents from MongoFB."""
        return self.db[collection].find(mongo_filter, *args, **kwargs)


class MongoMetricSource(MongoSource, MetricSource):
    """Gets metrics from MongoDB."""

    def _execute_query(self, metric_filter: JDict) -> MetricList:
        return [
            self._document_to_metric(document)
            for document in self.collection.find(metric_filter)
        ]

    @staticmethod
    def _document_to_metric(document: JDict) -> MetricInstance:
        del document["_id"]
        return Metric.from_dict(document)



"""Metric Store using MongoDB."""

from sinai.adaptors import Mongo
from sinai.exceptions import MetricNotFound
from sinai.metrics import Metric
from sinai.stores import MetricStore
from sinai.types import JDict, MetricId, MetricInstance, MetricResult


class MongoMetricStore(Mongo, MetricStore):
    """Metric store using MongoDB. Don't forget to set `database_name` and `connection_string`."""

    def _save(self, metric: MetricInstance) -> None:
        self.collection.insert_one(metric.pre_save(self.monitor.id))

    def _find_metric_by_filter(
        self, metric: MetricInstance, metric_filter: JDict
    ) -> MetricResult:
        if result := self.collection.find_one(metric_filter):
            mongo_id = result.pop("_id")
            existing_metric = Metric.from_dict(result)
            metric.created_at = existing_metric.created_at
            return metric, mongo_id
        raise MetricNotFound(f"Metric with fields: {metric_filter} not found.")

    def _replace(self, metric: MetricInstance, metric_id: MetricId) -> None:
        self.collection.replace_one(
            {"_id": metric_id}, metric.pre_save(self.monitor.id)
        )



"""A `Source` provides the input data to the Rule."""

from typing import Any

from sinai import BaseView
from sinai.adaptors import Memory
from sinai.types import JDict, MetricList


class Source(BaseView):
    """Any data source (base class)."""

    def get_data(self, request: JDict) -> Any:
        """Some sources can get their data at initialisation,
        others need to wait for the Rule to request it.
        This method is for that latter case.
        """


class MetricSource(Source, Memory):
    """In memory source of Metrics, use an adaptor instead when you need persistence."""

    def get(
        self,
        name: str = "",
        ref: str = "",
        value: str = "",
        context: str = "",
        monitor_id: str = "",
    ) -> MetricList:
        """Get a metric/metrics from the Source"""
        metric_filter = {}
        if name:
            metric_filter["name"] = name
        if ref:
            metric_filter["ref"] = ref
        if value:
            metric_filter["value"] = value
        if context:
            metric_filter["context"] = context
        if monitor_id:
            metric_filter["monitor_id"] = monitor_id
        return self._execute_query(metric_filter)

    def _execute_query(self, metric_filter: JDict) -> MetricList:
        results = self.find_metric(metric_filter)
        return list(results.values())



__all__ = ["Store", "MetricStore", "MongoMetricStore", "MemoryMetricStore"]
from sinai.stores.base import MetricStore, Store
from sinai.stores.memory import MemoryMetricStore
from sinai.stores.mongo import MongoMetricStore



__all__ = ["Source", "MetricSource", "MongoSource", "MongoMetricSource", "ApiSource"]

from sinai.sources.api import ApiSource
from sinai.sources.base import MetricSource, Source
from sinai.sources.mongo import MongoMetricSource, MongoSource



"""A Rule takes data from sources, makes metrics and puts them into stores."""
from __future__ import annotations

from sinai.types import Evaluation, MonitorInstance, SourceClasses, StoreClasses


class Rule:
    """The sources and stores are defined in the rule, instantiated by the monitor."""

    sources: SourceClasses = []
    stores: StoreClasses = []

    def __init__(self, monitor: MonitorInstance):
        self.monitor = monitor

    def evaluate(self) -> Evaluation:
        """The evalutation function is called by the monitor, which stores any returned metrics."""
        return None



from __future__ import annotations

from sinai.metrics import AGGREGATION_CLASSES
from sinai.rules.base import Rule
from sinai.types import (
    AggregationMetrics,
    Evaluation,
    MetricClasses,
    MetricList,
    MetricSourceClass,
    MetricSourceClasses,
    MetricSourceInstance,
    SList,
)


class MetricAggregationRule(Rule):
    """A rule that aggregates stored metrics and produces a metric."""

    sources: MetricSourceClasses = []
    update: SList = []
    count: MetricClasses = []
    sum: MetricClasses = []
    max: MetricClasses = []
    min: MetricClasses = []
    mean: MetricClasses = []
    mode: MetricClasses = []
    median: MetricClasses = []
    last_only = False

    def evaluate(self) -> Evaluation:
        self.metrics: AggregationMetrics = []
        for agg in AGGREGATION_CLASSES:
            met_cls = getattr(self, agg)
            if met_cls:
                base_name = "+".join([cls.name for cls in met_cls])
                self._process_agg(agg, base_name)
        return self.metrics

    def _process_agg(self, aggregation_name: str, base_name: str) -> None:
        metric_instances = []
        for source_cls in self.sources:
            metric_instances.extend(
                self._get_metric_instances(source_cls, aggregation_name)
            )
        data_points = [metric.value for metric in metric_instances]
        metric_cls = AGGREGATION_CLASSES[aggregation_name]
        metric = metric_cls(name=f"{base_name}:{metric_cls.name}", update=self.update)
        metric.process_data_points(data_points)
        self.metrics.append(metric)

    def _get_metric_instances(
        self, source_cls: MetricSourceClass, aggregation_name: str
    ) -> MetricList:
        metric_instances = []
        source: MetricSourceInstance = self.monitor.source(source_cls)  # type: ignore
        for target_class in getattr(self, aggregation_name):
            if self.last_only:
                instances = source.get(
                    name=target_class.name, monitor_id=self.monitor.id
                )
            else:
                instances = source.get(name=target_class.name)
            metric_instances.extend(instances)
        return metric_instances



"""Sinai is a library to help you monitor everything."""
from abc import ABC

from sinai.types import JDict, MonitorInstance

VERSION = "0.0.8"


class BaseView(ABC):
    def __init__(self, monitor: MonitorInstance):
        self.monitor = monitor
        self.data: JDict = {}  # Freeform working space to store your data.



"""The Api adaptors provide sources and stores using API endpoints.
Requires the requests library to be installed.
"""
from sinai.adaptors.api import ApiConnection
from sinai.exceptions import SourceError
from sinai.sources.base import Source
from sinai.types import JDict, MonitorInstance


class ApiSource(Source, ApiConnection):
    """Get data from an API endpoint."""

    def __init__(self, monitor: MonitorInstance):
        super().__init__(monitor)
        self.content: JDict = {}
        self.get()

    def get(self) -> None:
        """Get the required data from the specified API."""
        response = self.requests.get(self.url, headers=self.get_headers())
        if response.status_code == 200:
            self.content = response.json()
        else:
            raise SourceError(
                f"The URL {self.url} returned status {response.status_code}."
            )
